package org.kropp.index.impl;

import java.util.*;

import org.kropp.index.*;
import org.dbxml.core.data.Value;
import org.exist.storage.store.*;
import org.exist.util.*;
import org.dbxml.core.indexer.*;

public class SearchEngine {
	/**
	 * This is an utility class.
	 */
	private SearchEngine() {
	}

	/**
	 * Searches through given index for nodes satisfying given query.
	 * 
	 * @param index Index to search through
	 * @param query Query ti be used in search
	 * @return List of found strings.
	 */
	public static List<String> search(Index index, Query query) {
		ArrayList<String> result = new ArrayList<String>();

		BFile dataFile = index.getDataFile();

		Value[] indices = new Value[] { new Value(ByteConversion.intToByte(-1)) };

		for (Condition cond : query.getConditions()) {
			/*
			 * Evaluating next condition with respect to previously found query
			 * windows.
			 */
			indices = retrieveQueryWindow(index, indices, cond);

			/*
			 * There is no need to continue search if we didn't find anything.
			 */
			if (indices.length == 0) {
				return Collections.emptyList();
			}
		}

		/*
		 * Making result
		 */
		try {
			boolean checkAttribute = query.getConditions().get(query.getConditions().size()-1).getAxis() == Query.Axis.ATTRIBUTE;
			ArrayList resultData = dataFile.findValues(new IndexQuery(indices));
			for (Object node : resultData) {
				IndexNodeImpl indNode = IndexNodeImpl.fromValue((Value) node);
				
				/*
				 * Checking attribute flag.
				 */
				if(!checkAttribute || indNode.isAttribute()) {
					result.add(indNode.getNodeValue());
				}
			}
		} catch (Exception e) {
			e.printStackTrace();
		}

		return result;
	}

	private static Value[] retrieveQueryWindow(Index index, Value[] indices,
			Condition cond) {
		ArrayList resultNodes = new ArrayList();
		ArrayList<Value> newIndices = new ArrayList<Value>();

		try {
			if ((cond.getAxis() == Query.Axis.CHILD)
					|| (cond.getAxis() == Query.Axis.ATTRIBUTE)) {
				resultNodes = index.getParentIndex().findValues(
						new IndexQuery(indices));

				for (Object o : resultNodes) {
					int size = ((Value) o).getData().length / 4;
					for (int i = 0; i < size; i++) {
						byte[] val = new byte[4];
						System.arraycopy(((Value) o).getData(), i * 4, val, 0,
								4);
						newIndices.add(new Value(val));
					}
				}
			} else if (cond.getAxis() == Query.Axis.DESCENDANT) {
				for (Value preorder : indices) {
					resultNodes = index.getPostorderIndex().findValues(
							new IndexQuery(IndexQuery.GT,
									getPostorderFromPreorder(index, preorder)));

					for (Object o : resultNodes) {
						Value val = (Value) o;

						if (preorder.compareTo(val) < 0) {
							newIndices.add(val);
						}
					}
				}
			} else if (cond.getAxis() == Query.Axis.DESCENDANT_OR_SELF) {
				for (Value preorder : indices) {
					resultNodes = index.getPostorderIndex().findValues(
							new IndexQuery(IndexQuery.LEQ,
									getPostorderFromPreorder(index, preorder)));

					for (Object o : resultNodes) {
						Value val = (Value) o;
						if (preorder.compareTo(val) <= 0) {
							newIndices.add(val);
						}
					}
				}
			} else if (cond.getAxis() == Query.Axis.PARENT) {
				for (Value preorder : indices) {
					resultNodes = index.getPostorderIndex().findValues(
							new IndexQuery(IndexQuery.GT,
									getPostorderFromPreorder(index, preorder)));

					for (Object o : resultNodes) {
						Value val = (Value) o;
						if (getParentFromPreorder(index, preorder).compareTo(
								val) == 0) {
							newIndices.add(val);
						}
					}
				}
			} else if (cond.getAxis() == Query.Axis.ANCESTOR) {
				for (Value preorder : indices) {
					resultNodes = index.getPostorderIndex().findValues(
							new IndexQuery(IndexQuery.GT,
									getPostorderFromPreorder(index, preorder)));

					for (Object o : resultNodes) {
						Value val = (Value) o;
						if (preorder.compareTo(val) > 0) {
							newIndices.add(val);
						}
					}
				}
			} else if (cond.getAxis() == Query.Axis.ANCESTOR_OR_SELF) {
				for (Value preorder : indices) {
					resultNodes = index.getPostorderIndex().findValues(
							new IndexQuery(IndexQuery.GEQ,
									getPostorderFromPreorder(index, preorder)));

					for (Object o : resultNodes) {
						Value val = (Value) o;
						if (preorder.compareTo(val) >= 0) {
							newIndices.add(val);
						}
					}
				}
			} else if (cond.getAxis() == Query.Axis.FOLLOWING) {
				for (Value preorder : indices) {
					resultNodes = index.getPostorderIndex().findValues(
							new IndexQuery(IndexQuery.GT,
									getPostorderFromPreorder(index, preorder)));

					for (Object o : resultNodes) {
						Value val = (Value) o;
						if (preorder.compareTo(val) < 0) {
							newIndices.add(val);
						}
					}
				}
			} else if (cond.getAxis() == Query.Axis.PRECEDING) {
				for (Value preorder : indices) {
					resultNodes = index.getPostorderIndex().findValues(
							new IndexQuery(IndexQuery.LT,
									getPostorderFromPreorder(index, preorder)));

					for (Object o : resultNodes) {
						Value val = (Value) o;
						if (preorder.compareTo(val) > 0) {
							newIndices.add(val);
						}
					}
				}
			} else if (cond.getAxis() == Query.Axis.FOLLOWING_SIBLING) {
				for (Value preorder : indices) {
					resultNodes = index.getPostorderIndex().findValues(
							new IndexQuery(IndexQuery.GT,
									getPostorderFromPreorder(index, preorder)));

					for (Object o : resultNodes) {
						Value val = (Value) o;
						if (preorder.compareTo(val) < 0) {
							if (getParentFromPreorder(index, preorder)
									.compareTo(
											getParentFromPreorder(index, val)) == 0) {
								newIndices.add(val);
							}
						}
					}
				}
			} else if (cond.getAxis() == Query.Axis.PRECEDING_SIBLING) {
				for (Value preorder : indices) {
					resultNodes = index.getPostorderIndex().findValues(
							new IndexQuery(IndexQuery.LT,
									getPostorderFromPreorder(index, preorder)));

					for (Object o : resultNodes) {
						Value val = (Value) o;
						if (preorder.compareTo(val) > 0) {
							if (getParentFromPreorder(index, preorder)
									.compareTo(
											getParentFromPreorder(index, val)) == 0) {
								newIndices.add(val);
							}
						}
					}
				}
			} else {
				// Unknown axis
				return new Value[0];
			}
		} catch (Exception e) {
			e.printStackTrace();
		}

		if (newIndices.size() == 0) {
			return new Value[0];
		}
		indices = new Value[newIndices.size()];
		newIndices.toArray(indices);

		/*
		 * Filtering current subset by condition's nodeName.
		 */

		if ((cond.getName().length() > 0) && (!cond.getName().equals("*"))) {
			return filterByNodeName(index, indices, cond.getName());
		} else {
			if (newIndices.size() > 0) {
				indices = new Value[newIndices.size()];
				return newIndices.toArray(indices);
			} else {
				return new Value[0];
			}
		}
	}

	/**
	 * Returns postorder key of node identified by given preorder key.
	 * 
	 * @param index
	 *            Index that is used
	 * @param preorder
	 *            Preorder key
	 * @return Postorder key
	 */
	private static Value getPostorderFromPreorder(Index index, Value preorder) {
		try {
			ArrayList resultData = index.getDataFile().findValues(
					new IndexQuery(preorder));
			IndexNode indNode = IndexNodeImpl.fromValue((Value) resultData
					.get(0));
			return indNode.getPostorderKey();
		} catch (Exception e) {
		}

		return getPostorderFromPreorder(index, new Value(ByteConversion
				.intToByte(0)));
	}

	/**
	 * Returns parent key of node identified by given preorder key.
	 * 
	 * @param index
	 *            Index that is used
	 * @param preorder
	 *            Preorder key
	 * @return Parent key
	 */
	private static Value getParentFromPreorder(Index index, Value preorder) {
		try {
			ArrayList resultData = index.getDataFile().findValues(
					new IndexQuery(preorder));
			IndexNode indNode = IndexNodeImpl.fromValue((Value) resultData
					.get(0));
			return indNode.getParentKey();
		} catch (Exception e) {
		}

		return getParentFromPreorder(index, new Value(ByteConversion
				.intToByte(0)));
	}

	/**
	 * Filters nodes subset by given nodeName.
	 * 
	 * @param index Index used
	 * @param indices Nodes subset
	 * @param nodeName Name to filter by
	 * @return Filtered nodes subset
	 */
	private static Value[] filterByNodeName(Index index, Value[] indices, String nodeName) {
		ArrayList resultNodes = new ArrayList();
		ArrayList<Value> newIndices = new ArrayList<Value>();

		try {
			resultNodes = index.getPreorderIndex().findEntries(
					new IndexQuery(indices));
		} catch (Exception e) {
			e.printStackTrace();
		}

		for (Object o : resultNodes) {
			Value[] entry = (Value[]) o;

			int length = ByteConversion.byteToInt(entry[1].getData(), 0);
			String val = ByteConversion.UTF82String(entry[1].getData(), 4,
					length);

			if (val.equals(nodeName)) {
				newIndices.add(entry[0]);
			}
		}
		
		indices = new Value[newIndices.size()];
		return newIndices.toArray(indices);
	}
}
